"""
    experipy.exp
    ~~~~~~~~~~~~

    This module provides the Experiment class for running compositions 
    in the grammar, as well as the Exp Namespace for controlling and 
    configuring Experiment behavior.
"""
from __future__ import absolute_import

import shutil
import sys

from datetime   import datetime
from os         import chmod, makedirs, path
from subprocess import call
from time       import time

from .config    import Namespace
from .grammar   import Element
from .system    import cd, cp, mkdir, rm


Exp = Namespace("Exp",
    runsh   = "run.sh",
    shebang = "#!/bin/bash",
    rundir  = "/tmp",
    defname = "exp",
    out     = "raw.out",
    err     = "raw.err",
    timing  = "harness_time.out",
)


class Experiment(object):
    """Experiment objects perform the generation and execution of runscripts.
    
    Once a composition has been specified in the grammar, wrapping it in 
    an Experiment allows the user to generate a shell script as a string 
    using the make_runscript method. The run and queue methods provide 
    mechanisms for executing the generated scripts.

    Parameters
    ----------
    cmd : experipy.Element
        A composition of experipy Elements such as Executable and Group, 
        which defines the behavior the user wishes the Experiment to 
        perform.
    expname : str
        A name to be used for identifying the experiment. Defaults to 
        Exp.defname, which defaults to "exp".
    destdir : str
        An optional path to a directory where the results from running 
        the experiment should be stored. If None, expname will be used.
    """

    def __init__(self, cmd, expname=Exp.defname, destdir=None):
        if not isinstance(cmd, Element):
            raise TypeError("'{}' is not an instance of Element".format(cmd))

        self.cmd     = cmd
        self.expname = expname

        if not destdir:
            self.destdir = path.abspath(expname)
        else:
            self.destdir = path.abspath(destdir)
    

    def make_runscript(self, preamble=Exp.shebang, rm_rundir=True):
        """Create a string containing the experiment rendered as a shell script.

        Parameters
        ----------
        preamble : str
            The first line(s) of the runscript. Defaults to Exp.shebang, 
            which defaults to "#!/bin/bash".
        rm_rundir : bool
            If True, a line deleting the experiment's working directory 
            will be added to the end of the script. Defaults to True.

        Returns
        -------
        str
            A run script as described by the composition provided to the 
            Experiment.
        """

        # Name of the temporary directory for the experiment
        rundir = path.join(Exp.rundir, self.expname + "." + str(int(time())))
        
        # Start with the preamble
        scriptstr = preamble + "\n\n"
        
        # Collect experiment input files
        scriptstr += "# Experiment setup\n"

        scriptstr += str(mkdir(rundir, make_parents=True)) + "\n"
        scriptstr += str(cd(rundir)) + "\n"

        for infile in self.cmd.inputs():
            scriptstr += str(cp(path.abspath(infile), ".")) + "\n"
        
        # Execute the experiment components
        scriptstr += "\n# Run experiment\n"
        scriptstr += str(self.cmd) + "\n\n"
        
        # Exfill the experiment output files and clean up the rundir if needed
        scriptstr += "# Collect output files and clean up\n"
        for outfile in self.cmd.outputs():
            scriptstr += str(cp(outfile, self.destdir)) + "\n"
        
        if rm_rundir:
            scriptstr += str(rm(rundir)) + "\n"

        return scriptstr

    
    def run(self, rm_rundir=True):
        """Execute the experiment as a subprocess of the current process.

        Generates a run script, writes that script to the results 
        directory, and then executes the script as a subprocess of the 
        current process. The time the script takes to execute, including 
        setup and clean up time, is recorded. This function blocks until 
        the experiment is complete.

        Parameters
        ----------
        rm_rundir : bool
            If True, the directory created for running the experiment 
            will be deleted at the end of the experiment. Defaults to 
            True.
        """

        # Create the results directory, deleting any previous contents
        if path.exists(self.destdir):
            shutil.rmtree(self.destdir)
        makedirs(self.destdir)
        
        # Open the output, error and timing file handles for call
        out = open(path.join(self.destdir, Exp.out), 'w')
        err = open(path.join(self.destdir, Exp.err), 'w')
        timing = open(path.join(self.destdir, Exp.timing), 'w')

        # Write the runscript
        fname  = path.join(self.destdir, Exp.runsh)
        with open(fname, "w") as f:
            f.write(self.make_runscript(rm_rundir=rm_rundir))

        chmod(fname, 0o755)

        # Execute call and time the result
        start = datetime.now()
        call(fname, stdout=out, stderr=err)
        runtime = datetime.now() - start
        timing.write(str(runtime)+"\n")

        # Clean up and close file descriptors
        out.close()
        err.close()
        timing.close()
        

    def queue(self, h=False, n=False, q=None, A=None, **kwargs):
        """Submit the experiment to a job queuing system as a PBS script.
        
        Generates a script with a PBS script header, writes the script 
        to the results directory, and then submits it to the job queuing 
        system by running the command qsub as a subprocess.

        Parameters
        ----------
        h : bool
            Will add a ``-h`` to pbs headers if True, Default is False.
        n : bool
            Will add a ``-n`` to pbs headers if True, Default is False.
        q : str
            Optionally request a resource queue.
        A : str
            Optionally name the account to charge for this job.
        **kwargs : 
            The remaining keyword arguments will be combined into 
            resource requests with -l.
        """
       
        # Write the PBS script preamble
        pbsheader = (
            Exp.shebang + "\n#PBS -N {name}\n#PBS -o {qout}\n#PBS -e {qerr}"
        ).format(
            name=self.expname, 
            qout=path.join(self.destdir, Exp.out), 
            qerr=path.join(self.destdir, Exp.err)
        )
        if h == True:
            pbsheader += "\n#PBS -h"
        if n == True:
            pbsheader += "\n#PBS -n"
        if q != None:
            pbsheader += "\n#PBS -q " + q
        if A != None:
            pbsheader += "\n#PBS -A " + A
        if kwargs:
            reqs = ",".join(["{0}={1}".format(k, v) for k,v in kwargs.items()])
            pbsheader += "\n#PBS -l " + reqs

        # Create the results directory, deleting any previous contents
        if path.exists(self.destdir):
            shutil.rmtree(self.destdir)
        makedirs(self.destdir)
        
        # Write the runscript
        fname  = path.join(self.destdir, Exp.runsh)
        with open(fname, "w") as f:
            f.write(self.make_runscript(preamble=pbsheader))
        
        chmod(fname, 0o755)
        
        # Submit to the queue
        call(["qsub", fname])



from __future__ import absolute_import

from . import (
    config,
    exp,
    grammar,
    metrics,
    system,
)


__version__     = "0.3.1"

__title__       = "experipy"
__description__ = "A framework for writing and running Computational Science experiments"
__uri__         = "https://experipy.readthedocs.io"

__author__      = "Adam Howard"
__email__       = "ahoward0920@gmail.com"

__license__     = "BSD 3-clause"
__copyright__   = "Copyright (c) 2016 Adam Howard"


Experiment  = exp.Experiment
Namespace   = config.Namespace

dump_config = config.Namespace.dump_full_config

__all__ = [
    "config",
    "dump_config",
    "exp",
    "grammar",
    "metrics",
    "system",
    "Experiment",
    "Namespace",
]



"""
    experipy.grammar
    ~~~~~~~~~~~~~~~~

    This module provides the core elements which compose the Experipy 
    grammar: Executables, Wrappers, Pipelines, and Groups. These 
    elements facilitate specifying programs to execute as well as the 
    files they depend on. 
"""
import re

from .config import Namespace


class Element(object):
    """The Element class forms the grammar's base class.

    Parameters
    ----------
    inputs : list
        A list of strings which are the names of files that the Element 
        relies on for input. These will be copied to the run directory 
        when an Experiment is used to run the Element.
    outputs: list
        A list of strings which are the names of files that the Element 
        is expected to generate as output. These will be copied from the 
        run directory when an Experiment is used to run the Element.
    """

    def __init__(self, inputs=None, outputs=None):
        self._inputs = inputs if inputs else []
        self._outputs = outputs if outputs else []

    def inputs(self):
        """Generator which yields the Element's input files"""
        for item in self._inputs:
            yield item

    def outputs(self):
        """Generator which yields the Element's output files"""
        for item in self._outputs:
            yield item


class Executable(Element):
    """Executable objects should represent a single program and its arguments.

    Parameters
    ----------
    prog : str
        The name of the program executable.
    opts : list
        A list of command line options to pass to the program. Defaults to
        an empty list if not provided.
    wait : bool
        If False, a '&' will be appended to the argument list, indicating to
        the shell that it should background the program instead of blocking
        on it. Defaults to True.
    """

    def __init__(self, prog, opts=None, wait=True, **kwargs):
        super(Executable, self).__init__(**kwargs)
        self.prog = prog
        self.opts = []
        if opts != None:
            self.opts.extend(opts)

        if wait == False:
            self.opts.append("&")

    def __str__(self):
        """Render the Executable as it will appear in the shell script."""
        return "{0} {1}".format(self.prog, " ".join(map(str,self.opts)))


class Wrapper(Executable):
    """Wrapper objects allow specification of a program which wraps another.

    Wrappers are a subclass of Executable which allow specification of 
    programs such as GDB or Valgrind, which wrap around another program 
    to alter or observe its execution.

    Parameters
    ----------
    prog : str
        The name of the program executable.
    opts : list
        A list of command line options to pass to the program. Must 
        minimally contain a string having the value '[[wrapped]]', which 
        indicates where the wrapped executable should be inserted into 
        the wrapping executable's argument list.
    wrapped : experipy.Executable
        The wrapped Executable. Inputs and outputs specified to wrapped 
        will be included in the resultant object's inputs and outputs.
    wait : bool
        If False, a '&' will be appended to the argument list, 
        indicating to the shell that it should background the program 
        instead of blocking on it. Defaults to True.
    """

    def __init__(self, prog, opts, wrapped, **kwargs):
        if not isinstance(wrapped, Executable):
            raise TypeError("wrapped must be an instance of Executable")
        elif "[[wrapped]]" not in opts:
            raise ValueError("opts must contain '[[wrapped]]'")
        
        super(Wrapper, self).__init__(prog, opts, **kwargs)
        self.wrapped = wrapped

    def __str__(self):
        """Render the Wrapper as it will appear in the shell script."""
        return super(Wrapper, self).__str__().replace(
            "[[wrapped]]", str(self.wrapped)
        )
    
    def inputs(self):
        """Generator which yields the Wrapper's input files"""
        seen = set()
        for item in super(Wrapper, self).inputs():
            if item not in seen:
                seen.add(item)
                yield item
        for item in self.wrapped.inputs():
            if item not in seen:
                seen.add(item)
                yield item
    
    def outputs(self):
        """Generator which yields the Wrapper's output files"""
        seen = set()
        for item in super(Wrapper, self).outputs():
            if item not in seen:
                seen.add(item)
                yield item
        for item in self.wrapped.outputs():
            if item not in seen:
                seen.add(item)
                yield item


class Pipeline(Element):
    """Pipeline objects allow specification of pipelined workflows.

    A Pipeline takes one or more Element parts, and joins them with a 
    '|' operator, indicating to the shell that each part should recieve 
    its input from the previous part, and provide its output to the 
    next.

    Parameters
    ----------
    *parts : 
        One or more Executables or Wrappers to be chained together into 
        a pipeline. Inputs and outputs to the individual parts will be 
        included in the Pipeline's inputs and outputs.
    """

    def __init__(self, *parts, **kwargs):
        for part in parts:
            if not isinstance(part, Executable):
                raise TypeError("'{}' is not an instance of Executable".format(part))
       
        super(Pipeline, self).__init__(**kwargs)
        self.parts = parts

    def __str__(self):
        """Render the Pipeline as it will appear in the shell script."""
        return " | ".join(map(str,self.parts))
    
    def inputs(self):
        """Generator which yields the Pipeline's input files"""
        seen = set()
        for item in super(Pipeline, self).inputs():
            if item not in seen:
                seen.add(item)
                yield item
        for part in self.parts:
            for item in part.inputs():
                if item not in seen:
                    seen.add(item)
                    yield item
    
    def outputs(self):
        """Generator which yields the Pipeline's output files"""
        seen = set()
        for item in super(Pipeline, self).outputs():
            if item not in seen:
                seen.add(item)
                yield item
        for part in self.parts:
            for item in part.outputs():
                if item not in seen:
                    seen.add(item)
                    yield item


class Group(Element):
    """Group objects allow specification of Executables to be run in order.

    In the resultant script, a Group's parts will be included one after 
    another, in the order they were specified. Groups should be used 
    when specifying complex experiments involving multiple steps like 
    set up or post-processing, or combined with the wait parameter to 
    Executable to specify programs which should be run concurrently. A
    Group can also be used as a part in another Group.

    Parameters
    ----------
    *parts : 
        One or more Elements to be placed into the script. Inputs and 
        outputs to the individual parts will be included in the Group's
        inputs and outputs.
    """

    def __init__(self, *parts, **kwargs):
        for part in parts:
            if not (isinstance(part, Element)):
                raise TypeError("'{}' is not an instance of Element".format(part))

        super(Group, self).__init__(**kwargs)
        self.parts = parts

    def __str__(self):
        """Render the Group as it will appear in the shell script."""
        return "\n".join(map(str,self.parts))

    def inputs(self):
        """Generator which yields the Group's input files"""
        seen = set()
        for item in super(Group, self).inputs():
            if item not in seen:
                seen.add(item)
                yield item
        for part in self.parts:
            for item in part.inputs():
                if item not in seen:
                    seen.add(item)
                    yield item
    
    def outputs(self):
        """Generator which yields the Group's output files"""
        seen = set()
        for item in super(Group, self).outputs():
            if item not in seen:
                seen.add(item)
                yield item
        for part in self.parts:
            for item in part.outputs():
                if item not in seen:
                    seen.add(item)
                    yield item




import os
import shutil
import unittest

from experipy.exp       import Exp, Experiment
from experipy.grammar   import Executable

class TestExperimentInternals(unittest.TestCase):
    def test_make_runscript(self):
        exp = Experiment(Executable("hostname"))
        script = exp.make_runscript()
        self.assertIn(Exp.shebang, script)
        self.assertIn("mkdir", script)
        self.assertIn("hostname", script)
        self.assertIn("rm", script)

class TestExperimentRun(unittest.TestCase):
    def tearDown(self):
        shutil.rmtree("results")

    def test_simple_run(self):
        exp = Experiment(
            Executable("echo", ["Hello", "> test.out"], outputs=["test.out"]),
            destdir="results"
        ).run()
        self.assertTrue(os.path.exists("results"))
        self.assertTrue(os.path.exists("results/run.sh"))
        self.assertTrue(os.path.exists("results/test.out"))
        self.assertTrue(os.path.exists(os.path.join("results", Exp.timing)))



import unittest

from experipy.grammar import Executable, Wrapper, Pipeline, Group

class TestExecutable(unittest.TestCase):
    def test_render(self):
        exe = Executable("cat", ["1.txt", "2.txt"])
        self.assertEqual(str(exe), "cat 1.txt 2.txt")

    def test_output(self):
        exe = Executable("echo", 
            ['"Hello World"', "> test.txt"], 
            outputs=["test.txt"]
        )
        self.assertEqual(list(exe.outputs()),["test.txt"])

    def test_wait(self):
        exe = Executable("ping", wait=False)
        self.assertEqual(str(exe), "ping &")


class TestWrapper(unittest.TestCase):
    def test_render(self):
        exe = Executable("ls", ["-l"])
        wrap = Wrapper("valgrind", ["-v", "[[wrapped]]"], exe)
        self.assertEqual(str(wrap), "valgrind -v ls -l")

    def test_bad_opts_params(self):
        with self.assertRaises(ValueError):
            wrap = Wrapper("valgrind", ["-v"], Executable("ls"))

    def test_bad_wrapped(self):
        with self.assertRaises(TypeError):
            wrap = Wrapper("valgrind", ["-v", "[[wrapped]]"], dict())

    def test_inputs(self):
        exe = Executable("test", inputs=["1.txt", "3.txt"])
        wrap = Wrapper("wrap", ["--","[[wrapped]]"], exe, inputs=["2.txt","3.txt"])
        self.assertEqual(sorted(list(wrap.inputs())), ["1.txt", "2.txt", "3.txt"])



"""
    experipy.metrics
    ~~~~~~~~~~~~~~~~

    This module provides the Metric class as a means of defining and 
    extracting values from the results of Experiment runs.
"""
import os
import re


class Metric(object):
    """Metric objects define a value to be extracted from a given file.
    
    A metric consists of a base filename, a regex with which to search
    that file, and a parser which converts the value, once found, into 
    the desired type. 

    Parameters
    ----------
    name : str
        The name of the metric.
    filename : str
        The name of the file in a given results directory to search. 
        For instance, if the metric should appear in the standard 
        output of a given experiment, then filename should be set as
        ``raw.out``.
    regex : str
        A string which will be compiled as a regular expression and 
        used to search for the metric. Must contain a Named Group with 
        the name ``value`` (i.e. ``(?P<value>\d+)``).
    parser : callable
        A callable taking a single string argument and returning the
        value converted to the desired type. Defaults to ``float``.
    """
    def __init__(self, name, filename, regex, parser=float):
        self.name = name
        self.filename = filename
        self.parser = parser

        if not re.search(r"\(\?P<value>.*\)", regex):
            raise ValueError("regex must contain a 'value' named group")
        self.regex = re.compile(regex)


    def get_value(self, resultpath, default=None):
        """Given a path to a results directory, attempt to extract the 
        value. Optionally provide a default value in the event the value
        can't be found.
        """
        fname = os.path.join(resultpath, self.filename)
        with open(fname) as f:
            for line in f:
                m = re.search(self.regex, line)
                if m:
                    return self.parser(m.group('value'))
        return default



"""
    experipy.system
    ~~~~~~~~~~~~~~~

    This module provides a number of system and shell tools for helping to
    specify common tasks within the experipy grammar.
"""
from os         import path

from .grammar   import Executable

def cd(dirname):
    return Executable("cd", [dirname])


def cp(target, dest, opts=[]):
    return Executable("cp", opts + ['-R', '-t', dest, target])


def mkdir(dirname, make_parents=False):
    opts = []
    if make_parents:
        opts.append("--parents")
    return Executable("mkdir", opts + [dirname])


def mkfifo(pipename):
    return Executable("mkfifo", [pipename])


def mv(target, dest):
    return Executable("mv", [target, dest])


def rm(*files):
    return Executable("rm", ["-rf"] + list(files))


def wait():
    return Executable("wait")


def python_script(script, sopts=[], pythonexe="python", **kwargs):
    return Executable(pythonexe, [path.abspath(script)] + sopts, **kwargs)


def java_app(jarfile, popts=[], javaexe="java", jopts=[], **kwargs):
    jarfile = path.abspath(jarfile)
    return Executable(javaexe, jopts + ["-jar", jarfile] + popts, **kwargs)



"""
    experipy.config
    ~~~~~~~~~~~~~~~

    This module provides the Namespace class, which provides a mechanism 
    for defining collections of configurable constants.
"""
import os
import sys

if sys.version_info[0] == 2:
    import ConfigParser as configparser
else:
    import configparser


class Namespace(object):
    """Namespace objects are intended to act as collections of constants.

    All arguments passed to the Namespace when it is instantiated are 
    bound to attributes of the instance, allowing attribute reference as 
    opposed to dictionary access syntax. For example: 
    ``n = Namespace("N", foo="bar")`` would generate a namespace with an 
    attribute ``n.foo`` whose value is ``"bar"``. 

    Namespaces also support configuration using configparser INI files. 
    By default, configuration is stored and read from ``~/.experipyrc``, 
    unless the environment variable ``EXPERIPY_CONFIG_PATH`` is set, in 
    which case that value is used as the filename.

    Parameters
    ----------
    name : str
        The name to assign to the namespace. If not provided, the 
        resulting Namespace instance will be anonymous and not 
        configurable via the configuration file. 
    **kwargs :
        The remaining keyword arguments will be added the Namespace's
        dictionary, allowing for attribute access. If a name was 
        provided, and the namespace had a section in the configuration 
        file, conflicting arguments will have their values ignored in 
        favor of the value in the configuration file.
    """
    
    _config_file = os.getenv("EXPERIPY_CONFIG_PATH",
        os.path.expanduser("~/.experipyrc")
    )
    _config = configparser.RawConfigParser(allow_no_value=True)
    _config.read(_config_file)
    
    _registry = {}

    def __init__(self, name=None, **kwargs):
        self.attrs = dict(kwargs)
        self.name = name

        if name != None:
            self._register(name, self)

            # Override the defaults if they are specified in the config file
            if self._config.has_section(name):
                self.attrs.update(self._config.items(name))
    

    def __getattr__(self, name):
        try:
            return self.attrs[name]
        except KeyError:
            raise AttributeError("Namespace {} has no attribute: {}".format(
                self.name if self.name else "<unnamed>", name
            ))


    @classmethod
    def _register(cls, name, inst):
        """Record all named Namespaces who have been instantiated"""
        cls._registry[name] = inst


    @classmethod
    def dump_full_config(cls, fname=_config_file):
        """Write a config of all instantiated and preconfigured Namespaces.
        
        All instantiated and named Namespaces will be dumped to the 
        given file, along with any Namespace configurations which have 
        been loaded from the config, but whose corresponding Namespace 
        has not yet been instantiated.

        Parameters
        ----------
        fname : str
            Name of the file to write the config to. If not provided, it 
            will default to the current config file ("~/.experipyrc" or 
            the value of the EXPERIPY_CONFIG_PATH environment variable).
        """
        for name in cls._registry:
            if not cls._config.has_section(name):
                cls._config.add_section(name)

            ns = cls._registry[name]
            for attr in ns.attrs:
                cls._config.set(name, attr, ns.attrs[attr])

        with open(cls._config_file, 'w') as f:
            cls._config.write(f)


__all__ = [
    "Namespace",
]
